//+------------------------------------------------------------------+
//|                                                OrderWrappers.mq5 |
//|                                      Copyright  2010, Marketeer |
//|    v.2.0.6alpha, for comments and feedback visit mql4.com forums |
//+------------------------------------------------------------------+
#property copyright "Copyright  2010, Marketeer"
#property link      "http://www.mql4.com/"
#property library

//#include <stderror.mqh>

static int MaxRetries = 5;
static int TimeWaitMsecs = 5000;
static int SlippageEmulation = 0;

#include <marketeer\MarketMQL4.mqh>

//+------------------------------------------------------------------+
//| Error handling functions/wrappers                                |
//+------------------------------------------------------------------+
// invokes OrderSend, OrderClose, OrderModify, OrderDelete in cycles
// process errors
// use RefreshRates for selected errors


void SetMaxRetries(int number)
{
  MaxRetries = number;
}

void SetTimeWait(int milliseconds)
{
  TimeWaitMsecs = milliseconds;
}

void SetSlippageEmulation(int seconds) // maximum slippage
{
  SlippageEmulation = MathMax(seconds, 0);
}

static datetime SlipStart = 0;
static int SlipDuration = 0;

bool Slip()
{
  if(SlippageEmulation > 0)
  {
    if(SlipStart == 0)
    {
      SlipStart = TimeCurrent();
      SlipDuration = MathRand()*SlippageEmulation/32767;
      Print(TimeToString(TimeCurrent(), TIME_DATE | TIME_MINUTES | TIME_SECONDS), " Slippage emulation started for duration ", SlipDuration, " seconds");
      return(true);
    }
    if(TimeCurrent() - SlipStart < SlipDuration)
    {
      Print(TimeToString(TimeCurrent(), TIME_DATE | TIME_MINUTES | TIME_SECONDS), " Slippage emulation...");
      return(true);
    }
    else
    {
      Print(TimeToString(TimeCurrent(), TIME_DATE | TIME_MINUTES | TIME_SECONDS), " Slippage emulation ended");
      SlipStart = 0;
      SlipDuration = 0;
    }
  }
  return(false);
}


int HandledErrorCodes[] = //{0, ERR_SERVER_BUSY, ERR_NO_CONNECTION, ERR_TRADE_TIMEOUT, ERR_OFF_QUOTES, ERR_BROKER_BUSY, ERR_TRADE_CONTEXT_BUSY,
                          // ERR_INVALID_PRICE, ERR_PRICE_CHANGED, ERR_REQUOTE};
{TRADE_RETCODE_ERROR, TRADE_RETCODE_TIMEOUT, TRADE_RETCODE_ORDER_CHANGED, TRADE_RETCODE_TOO_MANY_REQUESTS, TRADE_RETCODE_CONNECTION,
 TRADE_RETCODE_REQUOTE, TRADE_RETCODE_PRICE_CHANGED, TRADE_RETCODE_PRICE_OFF};

int RequoteErrorCodes[] = //{ERR_INVALID_PRICE, ERR_PRICE_CHANGED, ERR_REQUOTE};
{TRADE_RETCODE_REQUOTE, TRADE_RETCODE_PRICE_CHANGED, TRADE_RETCODE_PRICE_OFF};


bool HandledError(string OpName, uint ErrCode)
{
  bool Result = false;
  for(int i = 0; i < ArraySize(HandledErrorCodes); i++)
  {
    if(HandledErrorCodes[i] == ErrCode) Result = true;
  }
  Print(OpName + " error: ", ErrCode, " result:", Result);
  return(Result);
}

bool PriceRequote(int ErrCode, int cmd)
{
  bool Result = false;
  for(int i = 0; i < ArraySize(RequoteErrorCodes); i++)
  {
    if(RequoteErrorCodes[i] == ErrCode) Result = true;
  }
  Print("Requote cmd: ", cmd, " result:", Result);
  return(Result);
}

#define OP_BUY       0
#define OP_SELL      1
#define OP_BUYLIMIT  2
#define OP_SELLLIMIT 3
#define OP_BUYSTOP   4
#define OP_SELLSTOP  5

int SuccessCodes[] = {TRADE_RETCODE_DONE, TRADE_RETCODE_PLACED, TRADE_RETCODE_CANCEL, TRADE_RETCODE_DONE_PARTIAL};

bool SuccessCode(uint retcode)
{
  bool Result = false;
  for(int i = 0; i < ArraySize(SuccessCodes); i++)
  {
    if(SuccessCodes[i] == retcode) Result = true;
  }
  //Print(OpName + " error: ", ErrCode, " result:", Result);
  return(Result);
}

void RefreshRates(string symbol)
{
  MqlTick tick;
  SymbolInfoTick(symbol, tick);
  SymbolInfoDouble(symbol, SYMBOL_BID);
}

int __LastError = 0;

int _GetLastError()
{
  return(__LastError);
}

int _OrderSend(string symbol, int cmd, double volume, double price, int slippage, double stoploss, double takeprofit,
                     string comment = "", int magic = 0, datetime expiration = 0, color arrow_color = CLR_NONE)
{
  MqlTradeRequest request;
  MqlTradeResult  result;
  bool Success;
  
  switch(cmd)
  {
    case OP_BUY:
      request.type = ORDER_TYPE_BUY;
    case OP_SELL:
      request.type = ORDER_TYPE_SELL;
      request.action = TRADE_ACTION_DEAL;
      break;
    case OP_BUYLIMIT:
      request.type = ORDER_TYPE_BUY_LIMIT;
    case OP_SELLLIMIT:
      request.type = ORDER_TYPE_SELL_LIMIT;
    case OP_BUYSTOP:
      request.type = ORDER_TYPE_BUY_STOP;
    case OP_SELLSTOP:
      request.type = ORDER_TYPE_SELL_STOP;
      request.action = TRADE_ACTION_PENDING;
      break;
  }
  
  //Print("volume=", DoubleToString(volume, 8), " price=", price);
  
  request.magic = magic;
  request.symbol = symbol;
  request.volume = volume;
  request.price = price;
  request.sl = stoploss;
  request.tp = takeprofit;
  request.deviation = slippage;
  request.type_filling = ORDER_FILLING_AON;
  request.type_time = ORDER_TIME_SPECIFIED;
  request.expiration = expiration;
  request.comment = comment;
  
  Success = OrderSend(request, result);
  
  __LastError = result.retcode;
  
  if(!Success || !SuccessCode(result.retcode))
  {
    if(result.retcode == TRADE_RETCODE_REQUOTE)
    {
      Print("Requote, new prices: ask=", result.ask, " bid=", result.bid);
    }
    return(-1);
  }
  
  int ticket = -1;
  
  if(request.action == TRADE_ACTION_DEAL)
  {
    long LID;
    Print("result.deal=", result.deal);
    bool bHistory = HistoryDealGetInteger(result.deal, DEAL_POSITION_ID, LID);
    if(!bHistory)
    {
      Print("HistoryDealGetInteger error:", GetLastError());
    }
    else
    {
      Print("LID=", LID);
    }
    ticket = LID;
  }
  else
  if(request.action == TRADE_ACTION_PENDING) ticket = result.order;
  
  return(ticket);
}

int OrderSendWrapper(string symbol, int cmd, double volume, double price, int slippage, double stoploss, double takeprofit,
                     string comment = "", int magic = 0, datetime expiration = 0, color arrow_color = CLR_NONE)
{
  int ticket = 0;
  for(int i = 0; i < MaxRetries; i++)
  {
    if(Slip())
    {
      ticket = -1; // last_error = 0
    }
    else
    {
      Print("_OrderSend price=", price);
      ticket = _OrderSend(symbol, cmd, volume, price, slippage, stoploss, takeprofit, comment, magic, expiration, arrow_color);
    }
    if(ticket == -1)
    {
      uint err = _GetLastError();
      if(!HandledError("Send", err)) break;
      Sleep(TimeWaitMsecs);
      RefreshRates(symbol);
      if(PriceRequote(err, cmd))
      {
        if(cmd == OP_BUY) price = MarketInfo(symbol, MODE_ASK);
        else if(cmd == OP_SELL) price = MarketInfo(symbol, MODE_BID);
        else break; // we don't know how to change price for orders that are not instant market orders
        Print("New price=", DoubleToString(price, MarketInfo(symbol, MODE_DIGITS)));
      }
    }
    else
    {
      break;
    }
  }
  return(ticket);
}

bool IsPosition(int ticket, string &PositionSymbol)
{
  long position;
  int n = PositionsTotal();
  for(int i = 0; i < n; i++)
  {
    PositionSymbol = PositionGetSymbol(i);
    if(PositionGetInteger(POSITION_IDENTIFIER, position))
    {
      if(position == ticket) return(true);
    } 
  }
  return(false);
}

string GetPositionSymbol(int ticket)
{
  long position;
  int n = PositionsTotal();
  for(int i = 0; i < n; i++)
  {
    string PositionSymbol = PositionGetSymbol(i);
    if(PositionGetInteger(POSITION_IDENTIFIER, position))
    {
      if(position == ticket) return(PositionSymbol);
    } 
  }
  return("");
}

bool _OrderModify(int ticket, double price, double stoploss, double takeprofit,
                        datetime expiration, color arrow_color = CLR_NONE)
{
  MqlTradeRequest request;
  MqlTradeResult  result;
  bool Success;
  string PositionSymbol;
  
  PositionSymbol = GetPositionSymbol(ticket);
  if(PositionSymbol != "")               // position
  {
    request.action = TRADE_ACTION_SLTP;
    request.symbol = PositionSymbol;      
  }
  else                                   // pending
  {
    if(!OrderSelect(ticket))
    {
      Print("Ticket ", ticket, " is neither order not position");
      return(false);
    }
    request.action = TRADE_ACTION_MODIFY;
    request.order = ticket;
    request.price = price;
    request.type_time = ORDER_TIME_SPECIFIED;
    request.expiration = expiration;
  }
  
  request.sl = stoploss;        // position & pending
  request.tp = takeprofit;      // position & pending

  Success = OrderSend(request, result);
  
  __LastError = result.retcode;
  
  if(!Success || !SuccessCode(result.retcode))
  {
    return(false);
  }
  
  return(true);
}

bool OrderModifyWrapper(int ticket, double price, double stoploss, double takeprofit,
                        datetime expiration, color arrow_color = CLR_NONE)
{
  bool Success = false;
  for(int i = 0; i < MaxRetries; i++)
  {
    if(Slip())
    {
      Success = false;
    }
    else
    {
      Success = _OrderModify(ticket, price, stoploss, takeprofit, expiration, arrow_color);
    }
    if(!Success)
    {
      if(!HandledError("Modify", _GetLastError())) break;
      Sleep(TimeWaitMsecs);
      string PositionSymbol = GetPositionSymbol(ticket);
      if(PositionSymbol != "") RefreshRates(PositionSymbol);
      else break;
    }
    else
    {
      break;
    }
  }
  return(Success);
}

bool _OrderCloseBy(int ticket, int opposite, color Color = CLR_NONE)
{
  return(false);
}

// is not possiblre in MT5
bool OrderCloseByWrapper(int ticket, int opposite, color Color = CLR_NONE)
{
  bool Success = false;
  for(int i = 0; i < MaxRetries; i++)
  {
    Success = _OrderCloseBy(ticket, opposite, Color);
    if(!Success)
    {
      int err = _GetLastError();
      if(!HandledError("CloseBy", err)) break;
      Sleep(TimeWaitMsecs);
      //RefreshRates();
    }
    else
    {
      break;
    }
  }
  return(Success);
}

int __PositionType = -1;

bool _OrderClose(int ticket, double lots, double price, int slippage, color Color = CLR_NONE)
{
  MqlTradeRequest request;
  MqlTradeResult  result;
  bool Success;
  string PositionSymbol;
  
  PositionSymbol = GetPositionSymbol(ticket);
  if(PositionSymbol == "")
  {
    Print("Ticket ", ticket, " is not a position");
    return(false);
  }
  
  request.action = TRADE_ACTION_DEAL;
  request.symbol = PositionSymbol;
  if(lots <= 0) lots = PositionGetDouble(POSITION_VOLUME);
  request.volume = lots;
  
  int ptype = PositionGetInteger(POSITION_TYPE);
  __PositionType = ptype;
  if(ptype == POSITION_TYPE_BUY)
  {
    ptype = POSITION_TYPE_SELL;
    request.price = SymbolInfoDouble(PositionSymbol, SYMBOL_BID);
  }
  else
  if(ptype == POSITION_TYPE_SELL)
  {
    ptype = POSITION_TYPE_BUY;
    request.price = SymbolInfoDouble(PositionSymbol, SYMBOL_ASK);
  }
  
  request.type = ptype;
  request.deviation = slippage;
  
  Success = OrderSend(request, result);
  
  __LastError = result.retcode;
  
  if(!Success || !SuccessCode(result.retcode))
  {
    return(false);
  }
  
  return(true);
}

int GetPositionType()
{
  return(__PositionType);
}

bool OrderCloseWrapper(int ticket, double lots, double price, int slippage, color Color = CLR_NONE)
{
  bool Success = false;
  for(int i = 0; i < MaxRetries; i++)
  {
    if(Slip())
    {
      Success = false;
    }
    else
    {
      Success = _OrderClose(ticket, lots, price, slippage, Color);
    }
    if(!Success)
    {
      int err = _GetLastError();
      if(!HandledError("Close", err)) break;
      Sleep(TimeWaitMsecs);
      string PositionSymbol = GetPositionSymbol(ticket);
      if(PositionSymbol != "") RefreshRates(PositionSymbol);
      else
      {
        Print("Can\'t select position #", ticket, " error:", err);
        break;
      }
      int cmd = GetPositionType();
      if(PriceRequote(err, cmd))
      {
        if(cmd == OP_BUY) price = MarketInfo(PositionSymbol, MODE_BID);
        else if(cmd == OP_SELL) price = MarketInfo(PositionSymbol, MODE_ASK);
        Print("New price=", DoubleToString(price, MarketInfo(PositionSymbol, MODE_DIGITS)));
      }
    }
    else
    {
      break;
    }
  }
  return(Success);
}

bool _OrderDelete(int ticket, color arrow_color = CLR_NONE)
{
  MqlTradeRequest request;
  MqlTradeResult  result;
  bool Success;

  request.action = TRADE_ACTION_REMOVE;
  request.order  = ticket;

  Success = OrderSend(request, result);
  
  __LastError = result.retcode;

  if(!Success || !SuccessCode(result.retcode))
  {
    return(false);
  }
  
  return(true);
}

bool OrderDeleteWrapper(int ticket, color arrow_color = CLR_NONE)
{
  bool Success = false;
  for(int i = 0; i < MaxRetries; i++)
  {
    if(Slip())
    {
      Success = false;
    }
    else
    {
      Success = _OrderDelete(ticket, arrow_color);
    }
    if(!Success)
    {
      if(!HandledError("Delete", _GetLastError())) break;
      Sleep(TimeWaitMsecs);
    }
    else
    {
      break;
    }
  }
  return(Success);
}

